open Format
open Formatx
open Listset
open Term
open Substitution
open Overlap
open Rewriting
open Iroiro
open Arg
open Result
open Monad.MonadList

(* proof tree *)

type proof = 
  | EMPTY  
  | KH12   of Rules.t (* R *) * Rules.t (* S *) * Rules.t (* CP_S(R) *)  * proof
  | HM11   of Rules.t (* R *) * Rules.t (* CPS'(R) *) * Rules.t (* CP(R) *)
  | HM13   of Rules.t (* R *) * Rules.t (* S *) * Rules.t (* CPS'(R,S) *)  * proof * proof
  | HM13CR of Rules.t (* R *) * Rules.t (* non-closed-CPS(R) *) 
  | OH14a  of Rules.t (* R *) * Rules.t (* CP(R) *) * Rules.t (* closing steps *) * proof
  | OH14b  of Rules.t (* R *) * Rules.t (* CPout(R) *) * Rules.t (* CPin(R) *) * Rules.t (* closing steps *) 
  | VO94   of Rules.t (* R *) * Rules.t (* CP(R) *) (* development closedness *)
  | AYT09  of Rules.t (* R *) * Rules.t (* CP(R) *) (* development closedness a la Toyama *)
  | VO08   of Rules.t (* R *) * (Rule.t * int) list (* rule label *)
  | JK86   of Rules.t (* R *) * Rules.t (* CP_AC(R) *)

let print_label ppf (rule, n) =
  fprintf ppf "[%a] = %d" Rule.print rule n 

let rec print_proof ppf = function
  | EMPTY ->
      fprintf ppf 
        "---- CR(R)@.\
         ---- R is empty.@."
  | HM11 (rs, cps, cp) ->
      fprintf ppf 
        "---- CR(R)@.\
         ---- [Hirokawa and Middeldorp, 2011]@.\
         R:@.%a@.\
         CPS'(R):@.%a@.\
         CP(R):@.%a@."
        Rules.print rs
        Rules.print cps
        Rules.print cp
  | KH12 (rs, ss, ecp, proof) ->
      fprintf ppf 
        "---- CR(R cup S)@.\
         ---- [Klein and Hirokawa, 2012]@.\
         R:@.%a@.\
         S:@.%a@.\
         CP_S(R):@.%a@.\
         CR(S):@.see below.@.@.%a"
        Rules.print rs Rules.print ss
        Rules.print ecp print_proof proof
  | HM13 (rs, ss, cps, proof1, proof2) ->
      fprintf ppf 
        "---- CR(R cup S)@.\
         ---- [Hirokawa and Middeldorp, 2013]@.\
         R:@.%a@.\
         S:@.%a@.\
         CPS(R,S):@.%a@.\
         CR(R):@.see below.@.\
         CR(S):@.see below.@.@.%a%a"
        Rules.print rs Rules.print ss
        Rules.print cps 
        print_proof proof1
        print_proof proof2
  | HM13CR (rs, cps) ->
      fprintf ppf 
        "---- CR(R)@.\
         ---- [Hirokawa and Middeldorp, 2013]@.\
         R:@.%a@.\
         non-closed-CPS(R):@.%a"
        Rules.print rs
        Rules.print cps 
  | OH14a (rs, cp, cs, proof) ->
      fprintf ppf 
        "---- CR(R)@.\
         ---- R : left-linear, C : CR and CP-closing, and Cdup/R : SN [Oyamaguchi-Hirokawa, 2014]@.\
         R:@.%a@.@.\
         CP(R):@.%a@.@.\
         critical-pair-closing system C:@.%a@.@.\
         CR(C): see below.@.@.%a@."
        Rules.print rs
        Rules.print cp
        Rules.print cs
        print_proof proof
  | OH14b (rs, cp_out, cp_in, cs) ->
      fprintf ppf 
        "---- CR(R)@.\
         ---- [Oyamaguchi-Hirokawa, 2014]@.\
         R:@.%a@.@.\
         outside critical pairs of R:@.%a@.@.\
         inside critical pairs of R:@.%a@.@.\
         critical-pair-closing system C:@.%a@.@."
        Rules.print rs
        Rules.print cp_out
        Rules.print cp_in
        Rules.print cs
  | JK86 (rs, cp) ->
      fprintf ppf 
        "---- CR(R)@.\
         ---- [Jouannaud and Kirchner, 1986]@.\
         R:@.%a@.\
         CP_AC(R):@.%a"
        Rules.print rs Rules.print cp
  | VO94 (rs, cp) ->
      fprintf ppf
        "---- CR(R)@.\
         ---- Development Closedness Theorem [van Oostrom, TCS 1994].@.\
         R:@.%a@.@.\
         CP(R):@.%a@.@."
        Rules.print rs Rules.print cp
  | AYT09 (rs, cp) ->
      fprintf ppf
        "---- CR(R)@.\
         ---- Extended Development Closedness Theorem [Aoto, Yoshida and Toyama, RTA 2009].@.\
         R:@.%a@.@.\
         CP(R):@.%a@.@."
        Rules.print rs Rules.print cp
  | VO08 (rs, a) ->
      fprintf ppf
        "---- CR(R)@.\
         ---- Rule Labeling [van Oostrom, RTA 2008].@.\
         R:@.%a@.@.\
         Label:@.%a@.@."
        Rules.print rs (print_list print_label "@.") a

let print_disproof ppf (rs, s, t) =
  fprintf ppf
    "---- Non-confluence check@.\
     R:@.%a@.@.\
     Unjoinable conversion:@.%a <->*@.%a@.@." 
    Rules.print rs
    Term.print s Term.print t

(* Auciliary functions *)

exception Failed
exception Disproved of Term.t * Term.t

let rec try_find f = function
  | [] -> raise Failed
  | x :: xs -> try f x with Failed -> try_find f xs

let rec pairwise_find p = function
  | [] -> raise Failed
  | x :: xs ->
      try 
        let y = List.find (p x) xs in (x, y) 
      with Not_found -> pairwise_find p xs

(* Global variables *)

let filenames  = ref []
let t          = ref None 
let tt         = ref "NaTT.exe -v:0"
(*let actt       = ref "muterm -i"*)
let smt        = ref "minismt -m"
let wcrk       = ref 5
let rlvk       = ref 5
let rlck       = ref 5

let rec tcap rs = function
  | V _ -> V (fresh ())
  | F (f, ts) -> 
      let us = List.map (tcap rs) ts in
      let u = F (f, us) in 
      if List.exists (fun (l, _) -> unifiable u l) rs then
        V (fresh ())
      else
        u

let commute1 n rs ss (s, t) = 
  let ss = reachable n rs [s]
  and ts = reachable n ss [t] in
  intersect ss ts

let commute n rs ss es =
  List.for_all (commute1 n rs ss) es

let joinable n rs es = commute n rs rs es

(** Confluence criteria *)

(* Knuth and Bendix' criterion (1970) *)
let kb70 rs = 
  let cp = cp rs in
  if not (Rsn.not_sn rs []) && 
     Sn.sn !tt rs [] = YES () &&
     List.for_all (fun (s, t) -> nf rs s = nf rs t) cp
  then
    KH12 (rs, [], cp, HM11 ([], [], []))
  else
    raise Failed

(* Hirokawa and Middeldorp's criterion (JAR 2011) *)
let hm11 rs = 
  let p (s, t) = s <> t in
  let cps = cps ~p rs in
  let cp = cp rs in
  if Rules.left_linear rs && 
     not (Rsn.not_sn cps rs) && 
     joinable !wcrk rs cp &&
     Sn.sn !tt cps rs = YES ()
  then
    HM11 (rs, cps, cp)
  else 
    raise Failed

(* Development closedness theorem (Van Oostrom, TCS 1994) *)
let vo94 rs = 
  if Rules.left_linear rs then 
     let p (s, t) = List.mem t (multistep rs s) in
     let cp = cp rs in
     if List.for_all p cp then
       VO94 (rs, cp)
     else 
       raise Failed
  else
    raise Failed

(* Extended development closedness theorem (Aoto, Yoshida and Toyama, RTA 2009) *)
let root_overlap (_, p, _, _) = p = []

let closed rs o = 
  let s, t = cp_of_overlap o in
  if root_overlap o then 
    intersect (multistep rs s) (reachable !wcrk rs [t])
  else
    List.mem t (multistep rs s)

let ayt09 rs = 
  if Rules.left_linear rs &&
     List.for_all (closed rs) (overlap rs) then 
    AYT09 (rs, cp rs)
  else
    raise Failed

(* Van Oostrom's rule labeling (RTA 2008) *)

let rlv rs = 
  match Rlv.solve ~tool:!smt !rlvk rs with
  | Some a -> VO08 (rs, a)
  | None -> raise Failed

let rlc rs =
  match Rl.solve  ~tool:!smt !rlck rs with
  | Some a -> VO08 (rs, a)
  | None -> raise Failed

(* Non-confluence check *)

let unjoinable_by_tcap rs s t = 
  not (Substitution.unifiable (tcap rs s) (tcap rs t))

(*
let unjoinable_by_closure k rs s t = 
  let ss1 = reachable k rs [s]
  and ts1 = reachable k rs [t] in
  let ss2 = reachable 1 rs ss1
  and ts2 = reachable 1 rs ts1 in
  Listset.subset ss2 ss1 &&
  Listset.subset ts2 ts1 && 
  not (intersect ss2 ts2)

let unjoinable k rs s t =
  unjoinable_by_tcap rs s t ||
  unjoinable_by_closure 1 rs s t
*)

let rec ground = function
  | V x -> F (x, [])
  | F (f, ts) ->
      F (f, List.map ground ts)

let critical_peak rs =
  let* (_, _), _, (l2, _), mu = overlap rs in
  [ ground (substitute mu l2) ]

let non_confluent k rs =
  let f t = pairwise_find (unjoinable_by_tcap rs) (reachable k rs [t]) in
  let revs = List.map (fun (x,y) -> y,x) rs in
  try_find f (critical_peak (rs @ revs))

(* Klein and Hirokawa's criterion (LPAR 2012) *)

(* joinability of S-overlap *)

let kh12_overlap_aux rule1 rule2 =
  let l1, r1 = Rule.rename rule1
  and l2, r2 = Rule.rename rule2 in
  let* p = function_positions l2 in
  guard (unifiable (ren l1) (ren (subterm_at p l2))) >>= fun () ->
  guard (p <> [] || not (Rule.variant (l1, r1) (l2, r2))) >>= fun () ->
  [ (l1, r1), p, (l2, r2), mgu (subterm_at p l2) l1 ] 

let kh12_overlap rs =
  let overlaps =
    let* a = rs in
    let* b = rs in
    kh12_overlap_aux a b in
  Listset.unique overlaps

let kh12_cp rs =
  List.map
    (fun o -> cp_of_overlap o)
    (kh12_overlap rs)

let kh12 cr rs ss =
  let ecp =
    try kh12_cp rs with Not_unifiable -> raise Failed in
  if sno rs ss && 
     joinable !wcrk (rs @ ss) ecp &&
     Sn.sn !tt rs ss = YES ()
  then
    KH12 (rs, ss, ecp, cr ss)
  else
    raise Failed

let kh12_auto cr rs =
  let f ss = kh12 cr (Listset.diff rs ss) ss in
  let sss =
    let* ss = Listx.power rs in
    guard (ss <> [] && not (Listset.subset rs ss)) >>= fun () ->
    [ss] in
  try_find f sss

(* Jouannaud and Kirchner's criterion (1986) *)

let jk86 rr =
  match Ac_rewriting.find_ac rr with
  | None -> raise Failed
  | Some (rs, ac_sym, ac, ac_all) ->
      let cps = Rules.flatten ac_sym
          (Ac_overlap.cp  ac_sym rs @
           Ac_overlap.cp2 ac_sym rs ac_all) in
      (*
      let cps_nf =
        [ Ac_rewriting.nf ac_sym rs s,
                   Ac_rewriting.nf ac_sym rs t 
        | s, t <- cps ] in
      let nj =
        [ s,t | s,t <- Variant.rename_rules cps_nf; 
                        not (Ac_rewriting.equal ac_sym s t) ] in
      match nj with
              | [] -> JK86 (rr, Variant.rename_rules cps)
              | (s, t) :: _ -> raise (Disproved (s, t))
         *)
      match Ac_cr.ac_jk86 rr rr with
          YES () -> JK86 (rr, Variant.rename_rules cps)
  | _      -> raise Failed

let jk86_maybe rr = 
  try jk86 rr with Disproved (_, _) -> raise Failed

(* Hirokawa and Middeldorp's criterion (2013) *)

(*
let hm13 cr rs ss =
  let p (s, t) = true (* s <> t *) in
  let union = Listx.unique (rs @ ss) in
  if Rules.left_linear rs && 
     Rules.left_linear ss &&
     commute !wcrk ss rs (mutual_cp rs ss) then
     let cps = mutual_cps ~p rs ss in
     if Sn.sn !tt cps union = YES then
       HM13 (rs, ss, cps, cr rs, cr ss)
     else 
       raise Failed
  else
   raise Failed

let hm13_auto cr rs =
  if Rules.left_linear rs then
    let f (rs1, rs2) = hm13 cr rs1 rs2 in
    let l = Listx.power rs in
    try_find f
      [ rs1, rs2
      | rs1 <- l; rs2 <- l;
        not (Listset.equal rs1 rs2);
        Listset.equal rs (rs1 @ rs2) ]
  else
    raise Failed

let hm13b_non_closed_overlap rs (s, t) =
  not (List.mem t (multistep rs s))

let hm13b_cps rs ss =
  Listx.unique
    (cps2 ~p:(hm13b_non_closed_overlap ss) rs ss @
     cps2 ~p:(hm13b_non_closed_overlap rs) ss rs)

let hm13b cr rs ss =
  let union = Listx.unique (rs @ ss) in
  if Rules.left_linear rs && 
     Rules.left_linear ss &&
     commute !wcrk ss rs (mutual_cp rs ss) then
     let cps = hm13b_cps rs ss in
     if Sn.sn !tt cps union = YES then
       if Listset.equal rs ss then
         HM13CR (rs, cps)
       else
         HM13 (rs, ss, cps, cr rs, cr ss)
     else 
       raise Failed
  else
   raise Failed

let hm13b_auto cr rs =
  if Rules.left_linear rs then
    let f (rs1, rs2) = hm13b cr rs1 rs2 in
    let l = Listx.power rs in
    try_find f
      [ rs1, rs2
      | rs1 <- l; rs2 <- l;
        not (Listset.equal rs1 rs2);
        Listset.equal rs (rs1 @ rs2) ]
  else
    raise Failed

let hm13b_cr rs =
  let cr rs = raise Failed in
  hm13b cr rs rs
*)

let hm13c_cps rs ss =
  let peaks1 = 
    let* o = overlap2 rs ss in 
    let (t, s, u) = critical_peak_of_overlap o in
    guard (not (intersect (multistep ss t) (reachable !wcrk rs [u]))) >>= fun () ->
    [(t, s, u)] in
  let peaks2 =
    let* o = overlap2 ss rs in
    guard (not (root_overlap o)) >>= fun () ->
    let (t,s,u) = critical_peak_of_overlap o in
    guard (not (List.mem u (multistep rs t))) >>= fun () ->
    [(t, s, u)]
  in
  Listx.concat_map (fun (t,s,u) -> [s, t; s, u]) (peaks1 @ peaks2)
  |> Listx.unique

let hm13c cr rs ss =
  let union = Listx.unique (rs @ ss) in
  if Rules.left_linear rs && 
     Rules.left_linear ss &&
     commute !wcrk ss rs (mutual_cp rs ss) then
     let cps = hm13c_cps rs ss in
     if Sn.sn !tt cps union = YES () then
       if Listset.equal rs ss then
         HM13CR (rs, cps)
       else
         HM13 (rs, ss, cps, cr rs, cr ss)
     else 
       raise Failed
  else
   raise Failed

let hm13c_auto cr rs =
  if Rules.left_linear rs then
    let f (rs1, rs2) = hm13c cr rs1 rs2 in
    let l = Listx.power rs in
    let sub_systems =
      let* rs1 = l in
      let* rs2 = l in
      guard (not (Listset.equal rs1 rs2)) >>= fun () ->
      guard (Listset.equal rs (rs1 @ rs2)) >>= fun () ->
      [rs1, rs2] in
    try_find f sub_systems
  else
    raise Failed

let hm13 rs =
  let cr rs = raise Failed in
  hm13c cr rs rs

(* OH14a *)

let oh14a_closed cs (s, t) =
  intersect (reachable !wcrk cs [s]) (reachable !wcrk cs [t])

let oh14a cr rs cs =
  let cs_d =
    List.filter
      (fun rule -> not (Rule.is_non_duplicating rule))
      cs in
  eprintf "--@.%a@." Rules.print cs_d;
  if Rules.left_linear rs && Sn.sn !tt cs_d rs = YES () then
    let cps = cp rs in
    if cps = [] then
      OH14a (rs, [], [], cr [])
    else if List.for_all (oh14a_closed cs) cps then
      OH14a (rs, cps, cs, cr cs)
    else
      raise Failed
  else
   raise Failed

let oh14a_auto cr rs =
  if rs = [] then
    EMPTY
  else
    if Rules.left_linear rs then
      let css =
        List.filter
          (fun cs -> not (Listset.equal rs cs))
          (Listx.power rs) in
      try_find (oh14a cr rs) css
    else
      raise Failed

(* OH14b *)

let oh14b_outside_closed cs (s, t) = nf cs s = nf cs t

let oh14b_inside_closed cs (s, t) =
  eprintf "%a -||-> . *<- %a@." Term.print s Term.print t;
  intersect (parallel_step cs s) (reachable_all cs [t])

let is_overlay (_,p,_,_) = p = []

let cp_out_in rs =
  let os_out, os_in = List.partition is_overlay (Overlap.overlap rs) in
  (List.map Overlap.cp_of_overlap os_out,
   List.map Overlap.cp_of_overlap os_in)

let oh14b rs cs =
  eprintf "--@.%a@." Rules.print cs;
  let cp_out, cp_in = cp_out_in rs in
  if Rules.left_linear rs &&
     Sn.sn !tt cs [] = YES () &&
     let _ = eprintf "out@." in
     List.for_all (oh14b_outside_closed cs) cp_out &&
     let _ = eprintf "in@." in
     List.for_all (oh14b_inside_closed cs) cp_in 
  then
    OH14b (rs, cp_out, cp_in, cs)
  else
    raise Failed

let oh14b_auto rs =
  if Rules.left_linear rs then
    try_find (oh14b rs) (Listx.power rs)
  else
    raise Failed

let rec confluent1 rs = 
  try_find (fun f -> f rs)
    [ kb70; hm13; rlv; rlc; jk86_maybe; kh12_auto confluent1]
    (* [ jk86_maybe; kh12_auto confluent1] *)

let confluent rs = 
  try_find (fun f -> f rs)
    [ kb70; hm13; rlv; rlc; jk86; kh12_auto confluent1]
    (* [ jk86; kh12_auto confluent1] *)

(*
let rec confluent rs = hm11 rs

let confluent rs = oh14b_auto rs
*)

(*
let rec confluent rs = 
  try_find (fun f -> f rs) [oh14b_auto; oh14a_auto confluent]
*)

(*
module M = Map.Make
  (struct
     type t = Rules.t
     let compare rs ss =
       Stdlib.compare
         (List.sort Stdlib.compare rs)
         (List.sort Stdlib.compare ss)
   end)

let memo = ref M.empty

let rec confluent rs = 
  try 
    M.find rs !memo
  with Not_found ->
    let b =
      try_find (fun f -> f rs)
        [ hm13c_cr; rlv; hm13c_auto confluent ] in
    (memo := M.add rs b !memo; b)
*)

let options = align [ 
  "-t",  Float  (fun x -> t := Some x), "<f> second timeout"; 
  "-tt", String (fun x -> tt := x),
     sprintf "<tool> termination tool (default: %s)" !tt;
(*
  "-actt", String (fun x -> actt := x),
     sprintf "<tool> AC termination tool (default: %s)" !actt;
*)
  "-smt", String (fun x -> smt := x),
     sprintf "<tool> QF_NIA SMT solver (default: %s)" !smt;
]

(* main *)
     
let short_usage = "Usage: collsaigawa [options] <file>\n"
let usage = "Confluence tool CoLL-Saigawa v1.6\n" ^ short_usage ^ "Options are:"

let () = 
  parse options (fun it -> filenames := !filenames @ [it] ) usage;
  if List.length !filenames <> 1 then
    (eprintf "%s%!" short_usage; exit 1)
  else
    let filename = List.hd !filenames in
    let f_timeout () = printf "TIMEOUT@." in
    let f_prove () =
      let rs = Read.read_trs filename in
      if not (Rules.variable_condition rs) then
        (eprintf "Cannot handle TRS with extra variables.@."; exit 1);
      (* CoLL config *)
      Ref.confluence_flag_on ();
      if Rules.left_linear rs then 
        (* run coll with confluence mode *)
        Coll_main.run_cr [filename]
      else
        let _,rs1 = Trans.reduce !Ref.k rs in
        try 
          let s, t = non_confluent !wcrk rs1 in
          printf "NO@.@.%a@." print_disproof (rs1, s, t)
        with Failed ->
          try 
            let proof = confluent rs1 in
            printf "YES@.@.%a@." print_proof proof;
            Trans.is_transformed rs rs1;
          with
          | Disproved (s, t) -> printf "NO@.@.%a@." print_disproof (rs, s, t)
          | Failed -> printf "MAYBE@."
    in
    Alarm.try_with_timeout ?timeout:!t f_timeout f_prove ()
