(* Rule labeling based on conversions *)
open Minismt
open Term
open Substitution
open Overlap
open Wcr
open Format
open Monad.MonadList

let minimal ~geq xs = 
  let ys =
    List.filter
      (fun x ->
        List.for_all (fun y -> not (geq x y) || x = y) xs)
      xs in
  Listx.unique ys

(*
let minimal ~geq xs = xs
*)

let rec emb xs ys =
  match xs, ys with
  | _, [] -> true
  | [], _ :: _ -> false
  | x :: xs, y :: ys -> (x = y && emb xs ys) || emb xs (y :: ys)

let pairwise_emb (xs, ys) (xs', ys') = emb xs xs' && emb ys ys'

let rec lookup rule = function
  | [] -> failwith "Rlc.lookup"
  | (rule', x) :: _ when Rule.variant rule rule' -> x
  | _ :: a -> lookup rule a

let rec rewrite_root ix s = 
  let* ((l,r),x) = ix in
  guard (is_instance_of s l) >>= fun () ->
  [x, substitute (pattern_match l s) r]

let rewrite ix s =
  let xs =
    let* p = positions s in
    let* (x, t) = rewrite_root ix (subterm_at p s) in
    [x, replace s t p] in
  Listx.unique xs

(* X^(k) *)
let rec jj1 ix k ss =
  if k = 0 then ss else 
    let results =
      let* (xs,s) = ss in
      let* (x,t) = rewrite ix s in
      [xs @ [x], t] in
  minimal
    ~geq:(fun (xs1, s1) (xs2, s2) -> s1 = s2 && emb xs1 xs2) 
    (ss @ jj1 ix (k - 1) results)
	
let jj ix k (s, t) = 
  let l1 = jj1 ix k [[], s] 
  and l2 = jj1 ix k [[], t] in
  let joins =
    let* (xs,s') = l1 in
    let* (ys,t') = l2 in
    guard (s' = t') >>= fun () ->
    [xs,ys] in
  minimal ~geq:pairwise_emb joins

let psi ix rules a b = function
  | [] -> C []
  | c :: ds ->
      let finds =
        let* (rule,c') = ix in
        guard (List.mem rule rules) >>= fun () ->
        [c'] in
      let ord =
        if List.mem c finds then
          (fun x y -> Eq (x, y)) 
        else  
          (fun x y -> Gt (x, y))
      in
      let cnf = List.map (fun d -> D [Gt (a, d); Gt (b, d)]) ds
      in
      C (ord b c :: cnf)

let phi_ab ix rules a b cs =
  let dnf =
    let* (cs1, cs2) = List.combine (Listx.prefix cs) (Listx.suffix cs) in
    let fs = List.map (fun c -> Gt (a, c)) cs1 in
    [C (psi ix rules a b cs2 :: fs)] in
  D dnf

let phi ix rules k ((rule1, p, rule2, mu) as o) =
  let dnf =
    List.map
      (fun (cs,ds) ->
        C [phi_ab ix rules (lookup rule1 ix) (lookup rule2 ix) cs;
           phi_ab ix rules (lookup rule2 ix) (lookup rule1 ix) ds])
      (jj ix k (cp_of_overlap o)) in
  D dnf

let rl ix rules k = 
  C (List.map (phi ix rules k) (overlap rules))

(*
let inverse rules = 
  [ r, l | l, r <- rules; Listset.subset (variables l) (variables r) ]
*)
let inverse rules = 
  let* (l,r) = rules in
  guard (not (Term.is_variable r) &&
         Listset.subset (variables l) (variables r)) >>= fun () ->
  [r,l]

let rec unique ~eq = function
  | [] -> []
  | x :: ys ->
      let zs =
        List.filter
          (fun y -> not (eq x y))
          ys in
      x :: unique ~eq zs

let solve ~tool k rules =
  if Rules.linear rules then
    let inv = inverse rules in
    let union = unique ~eq:Rule.variant (Listset.union rules inv) in
    let ix =
      List.map
        (fun (n,rule) -> rule, sprintf "x%d" n)
        (Listx.ix ~i:1 union) in
    let eqs =
      List.map
        (fun (l,r) -> Eq (lookup (l, r) ix, lookup (r, l) ix))
        inv in
    match 
      Minismt.sat ~tool (List.map snd ix)
        (C (rl ix rules k :: eqs))
    with
    | Some b ->
        let map =
          let* (rule,x) = ix in
          guard (List.mem rule rules) >>= fun () ->
          [rule, List.assoc x b] in
        Some map
    | None -> None
  else 
    None
